// bibiman - a TUI for managing BibLaTeX databases
// Copyright (C) 2024  lukeflo
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
/////

use biblatex::{self, Bibliography};
use biblatex::{ChunksExt, Type};
use color_eyre::owo_colors::OwoColorize;
use itertools::Itertools;
use std::collections::HashMap;
use std::ffi::{OsStr, OsString};
use std::{fs, path::PathBuf};
use walkdir::WalkDir;

use crate::app;
use crate::cliargs::{self};
use crate::config::BibiConfig;

// Set necessary fields
// TODO: can surely be made more efficient/simpler
#[derive(Debug)]
pub struct BibiSetup {
    // pub bibfile: PathBuf,           // path to bibfile
    pub bibfilestring: String,      // content of bibfile as string
    pub bibliography: Bibliography, // parsed bibliography
    pub citekeys: Vec<String>,      // list of all citekeys
    pub keyword_list: Vec<String>,  // list of all available keywords
    pub entry_list: Vec<BibiData>,  // List of all entries
}

#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
pub struct BibiData {
    pub id: u32,
    pub authors: String,
    pub short_author: String,
    pub title: String,
    pub year: String,
    pub pubtype: String,
    pub keywords: String,
    pub citekey: String,
    pub abstract_text: String,
    pub doi_url: Option<String>,
    pub filepath: Option<Vec<OsString>>,
    pub file_field: bool,
    pub subtitle: Option<String>,
}

impl BibiData {
    // This functions decides which fields are rendered in the entry table
    // Fields which should be usable but not visible can be left out
    pub fn ref_vec(&mut self) -> Vec<&str> {
        self.short_author = match self.authors.split_once(",") {
            Some((first, _rest)) => {
                if self.authors().contains("(ed.)") {
                    let first_author = format!("{} et al. (ed.)", first);
                    first_author
                } else {
                    let first_author = format!("{} et al.", first);
                    first_author
                }
            }
            None => String::from(""),
        };

        vec![
            {
                if self.short_author.is_empty() {
                    self.authors()
                } else {
                    &self.short_author
                }
            },
            self.title(),
            self.year(),
            self.pubtype(),
        ]
    }

    pub fn entry_id(&self) -> &u32 {
        &self.id
    }

    pub fn authors(&self) -> &str {
        &self.authors
    }

    pub fn title(&self) -> &str {
        &self.title
    }

    pub fn year(&self) -> &str {
        &self.year
    }

    pub fn pubtype(&self) -> &str {
        &self.pubtype
    }

    pub fn citekey(&self) -> &str {
        &self.citekey
    }

    pub fn doi_url(&self) -> &str {
        self.doi_url.as_ref().unwrap()
    }

    // Gets the path of the associated file for a bib entry. If one is set explicitly
    // as a field, use that. If not, try to match it to one of the files found in the
    // pdf_files dir.
    pub fn filepath(&mut self) -> Vec<&OsStr> {
        self.filepath
            .as_mut()
            .unwrap()
            .iter_mut()
            .map(|p| p.as_os_str())
            .collect_vec()
    }

    pub fn subtitle(&self) -> &str {
        self.subtitle.as_ref().unwrap()
    }
}

impl BibiSetup {
    pub fn new(main_bibfiles: &[PathBuf], cfg: &BibiConfig) -> Self {
        // TODO: Needs check for config file path as soon as config file is impl
        Self::check_files(main_bibfiles);
        let bibfilestring = Self::bibfiles_to_string(main_bibfiles);
        let bibliography = biblatex::Bibliography::parse(&bibfilestring).unwrap();
        let citekeys = Self::get_citekeys(&bibliography);
        let keyword_list = Self::collect_tag_list(&citekeys, &bibliography);
        let entry_list = Self::create_entry_list(&citekeys, &bibliography, cfg);
        Self {
            // bibfile,
            bibfilestring,
            bibliography,
            citekeys,
            keyword_list,
            entry_list,
        }
    }

    // Check which file format the passed file has
    fn check_files(main_bibfiles: &[PathBuf]) {
        if main_bibfiles.is_empty() {
            println!(
                "{}",
                "No bibfile passed as argument or through config file. Please select a valid file."
                    .red()
                    .bold()
            );
            println!();
            println!("{}", cliargs::help_func());
            std::process::exit(1)
        } else {
            // Loop over all files and check for the correct extension
            main_bibfiles.iter().for_each(|f| {
                if f.extension().is_some() && f.extension().unwrap() != "bib" {
                    println!(
                        "{}\n{}",
                        "The passed file has no valid extension. You need a \'.bib\' file:"
                            .red()
                            .bold(),
                        f.as_os_str().to_string_lossy().bright_red().italic()
                    );
                    println!();
                    println!("{}", cliargs::help_func());
                    std::process::exit(1)
                }
            });
        }
    }

    fn bibfiles_to_string(main_bibfiles: &[PathBuf]) -> String {
        // Map PathBufs to String anc join the vector into one big string
        // This behaviour is needed by the biblatex crate for parsing
        let file_strings: Vec<String> = main_bibfiles
            .iter()
            .map(|f| fs::read_to_string(f).unwrap())
            .collect();

        file_strings.join("\n")
    }

    fn create_entry_list(
        citekeys: &[String],
        bibliography: &Bibliography,
        cfg: &BibiConfig,
    ) -> Vec<BibiData> {
        let mut pdf_files = if cfg.general.pdf_path.is_some() {
            collect_pdf_file_paths(cfg.general.pdf_path.as_ref().unwrap())
        } else {
            None
        };
        citekeys
            .iter()
            .enumerate()
            .map(|(i, k)| {
                let filepaths: (Option<Vec<OsString>>, bool) =
                    { Self::get_filepath(k, bibliography, &mut pdf_files) };

                BibiData {
                    id: i as u32,
                    authors: Self::get_authors(k, bibliography),
                    short_author: String::new(),
                    title: Self::get_title(k, bibliography),
                    year: Self::get_year(k, bibliography),
                    pubtype: Self::get_pubtype(k, bibliography),
                    keywords: Self::get_keywords(k, bibliography),
                    citekey: k.to_owned(),
                    abstract_text: Self::get_abstract(k, bibliography),
                    doi_url: Self::get_weblink(k, bibliography),
                    filepath: filepaths.0,
                    file_field: filepaths.1,
                    subtitle: Self::get_subtitle(k, bibliography),
                }
            })
            .collect()
    }

    // get list of citekeys from the given bibfile
    // this list is the base for further operations on the bibentries
    // since it is the entry point of the biblatex crate.
    pub fn get_citekeys(bibstring: &Bibliography) -> Vec<String> {
        let citekeys: Vec<String> = bibstring.keys().map(|k| k.to_owned()).collect();
        citekeys
    }

    // collect all keywords present in the bibliography
    // sort them and remove duplicates
    // this list is for fast filtering entries by topics/keyowrds
    pub fn collect_tag_list(citekeys: &[String], biblio: &Bibliography) -> Vec<String> {
        // Initialize vector collecting all keywords
        let mut keyword_list = vec![];

        // Loop over entries and collect all keywords
        for i in citekeys {
            if biblio.get(i).unwrap().keywords().is_ok() {
                let items = biblio.get(i).unwrap().keywords().unwrap().format_verbatim();
                // Split keyword string into slices, trim leading and trailing
                // whitespaces, remove empty slices, and collect them
                let mut key_vec: Vec<String> = items
                    .split(',')
                    .map(|s| s.trim().to_string())
                    .filter(|s| !s.is_empty())
                    .collect();
                // Append keywords to vector
                keyword_list.append(&mut key_vec);
            }
        }

        // Sort the vector and remove duplicates
        // keyword_list.sort_by(|a, b| a.to_lowercase().cmp(&b.to_lowercase()));
        keyword_list.sort_by_key(|a| a.to_lowercase());
        keyword_list.dedup();
        keyword_list
    }

    pub fn get_authors(citekey: &str, biblio: &Bibliography) -> String {
        if biblio.get(citekey).unwrap().author().is_ok() {
            let authors = biblio.get(citekey).unwrap().author().unwrap();
            if authors.len() > 1 {
                authors.iter().map(|a| &a.name).join(", ")
            } else if authors.len() == 1 {
                authors[0].name.to_string()
            } else {
                "empty".to_string()
            }
        } else if !biblio.get(citekey).unwrap().editors().unwrap().is_empty() {
            let editors = biblio.get(citekey).unwrap().editors().unwrap();
            if editors[0].0.len() > 1 {
                format!("{} (ed.)", editors[0].0.iter().map(|e| &e.name).join(", "))
            } else if editors[0].0.len() == 1 {
                format!("{} (ed.)", editors[0].0[0].name)
            } else {
                "empty".to_string()
            }
        } else {
            "empty".to_string()
        }
    }

    pub fn get_title(citekey: &str, biblio: &Bibliography) -> String {
        if biblio.get(citekey).unwrap().title().is_ok() {
            biblio
                .get(citekey)
                .unwrap()
                .title()
                .unwrap()
                .format_verbatim()
        } else {
            "no title".to_string()
        }
    }

    pub fn get_year(citekey: &str, biblio: &Bibliography) -> String {
        let bib = biblio.get(citekey).unwrap();
        if bib.date().is_ok() {
            let year = bib.date().unwrap().to_chunks().format_verbatim();
            year[..4].to_string()
        } else {
            "n.d.".to_string()
        }
    }

    pub fn get_pubtype(citekey: &str, biblio: &Bibliography) -> String {
        biblio.get(citekey).unwrap().entry_type.to_string()
    }

    pub fn get_keywords(citekey: &str, biblio: &Bibliography) -> String {
        if biblio.get(citekey).unwrap().keywords().is_ok() {
            biblio
                .get(citekey)
                .unwrap()
                .keywords()
                .unwrap()
                .format_verbatim()
        } else {
            "".to_string()
        }
    }

    pub fn get_abstract(citekey: &str, biblio: &Bibliography) -> String {
        if biblio.get(citekey).unwrap().abstract_().is_ok() {
            biblio
                .get(citekey)
                .unwrap()
                .abstract_()
                .unwrap()
                .format_verbatim()
        } else {
            "no abstract".to_string()
        }
    }

    pub fn get_weblink(citekey: &str, biblio: &Bibliography) -> Option<String> {
        let bib = biblio.get(citekey).unwrap();
        if bib.doi().is_ok() {
            Some(bib.doi().unwrap().trim().into())
        } else if bib.url().is_ok() {
            Some(bib.url().unwrap().trim().into())
        } else {
            None
        }
    }

    pub fn get_filepath(
        citekey: &str,
        biblio: &Bibliography,
        pdf_files: &mut Option<HashMap<String, Vec<PathBuf>>>,
    ) -> (Option<Vec<OsString>>, bool) {
        if biblio.get(citekey).unwrap().file().is_ok() {
            (
                Some(vec![biblio
                    .get(citekey)
                    .unwrap()
                    .file()
                    .unwrap()
                    .trim()
                    .into()]),
                true,
            )
        } else if pdf_files.is_some() {
            (Self::merge_filepath_or_none(&citekey, pdf_files), false)
        } else {
            (None, false)
        }
    }

    pub fn get_subtitle(citekey: &str, biblio: &Bibliography) -> Option<String> {
        if biblio.get(citekey).unwrap().subtitle().is_ok() {
            Some(
                biblio
                    .get(citekey)
                    .unwrap()
                    .subtitle()
                    .unwrap()
                    .format_verbatim(),
            )
        } else {
            None
        }
    }

    fn merge_filepath_or_none(
        citekey: &str,
        pdf_files: &mut Option<HashMap<String, Vec<PathBuf>>>,
    ) -> Option<Vec<OsString>> {
        let pdf_file = {
            // let mut idx = 0;
            let citekey = citekey.to_owned().to_ascii_lowercase() + ".pdf";
            // let filename = citekey.to_owned() + ".pdf";
            // for f in args.pdf_files.unwrap().iter() {
            //     if f.file_name().unwrap().to_str().unwrap() == &filename {
            //         break f;
            //     }
            // }

            // loop {
            //     if idx + 1 > pdf_files.as_ref().unwrap().len() {
            //         break None;
            //     }
            //     let cur_entry = pdf_files.as_ref().unwrap()[idx].clone();
            //     if cur_entry.is_file()
            //         && cur_entry
            //             .file_name()
            //             .unwrap()
            //             .to_ascii_lowercase()
            //             .to_str()
            //             .unwrap()
            //             == citekey
            //     {
            //         let path = cur_entry.to_owned().into_os_string();
            //         pdf_files.as_mut().unwrap().swap_remove(idx);
            //         break Some(path);
            //     } else {
            //         idx += 1
            //     }
            // }

            // for file in pdf_files.as_ref().unwrap().iter() {
            //     let filename = file.file_name().unwrap().to_ascii_lowercase();
            //     if filename.to_str().unwrap() == citekey {
            //         break;
            //     } else if pdf_files.as_ref().unwrap().len() > idx {
            //         break;
            //     } else {
            //         idx += 1;
            //     }
            // }

            // if pdf_files.as_ref().unwrap()[idx].is_file() {
            //     Some(pdf_files.as_ref().unwrap()[idx].to_owned().into_os_string())
            // } else {
            //     None
            // }

            if pdf_files.as_ref().unwrap().contains_key(&citekey) {
                let path_vec = pdf_files
                    .as_ref()
                    .unwrap()
                    .get(&citekey)
                    .unwrap()
                    .to_owned();
                Some(path_vec.into_iter().map(|p| p.into_os_string()).collect())
            } else {
                None
            }
        };

        pdf_file
    }
}

/// This function walks the given dir and collects all pdf files into a `HashMap`
/// of the format `[String, Vec<PathBuf>]`, where `String` represents the basename
/// of the file and the `Vec<PathBuf>` holds all filepaths ending with this basename.
///
/// In most cases the latter is only a single path, but there might be some concepts
/// with subdirs were some entries have multiple files associated with them.
pub fn collect_pdf_file_paths(pdf_dir: &PathBuf) -> Option<HashMap<String, Vec<PathBuf>>> {
    let mut files: HashMap<String, Vec<PathBuf>> = HashMap::new();

    // Expand tilde to /home/user
    let pdf_dir = if pdf_dir.starts_with("~") {
        &app::expand_home(&pdf_dir)
    } else {
        pdf_dir
    };

    // Walk the passed dir and collect all pdf files into hashmap
    if pdf_dir.is_dir() {
        for file in WalkDir::new(pdf_dir) {
            let f = file.unwrap().into_path();
            if f.is_file()
                && f.extension().is_some()
                && f.extension().unwrap_or_default().to_ascii_lowercase() == "pdf"
            {
                let filename = f
                    .file_name()
                    .unwrap()
                    .to_ascii_lowercase()
                    .into_string()
                    .unwrap();

                if let Some(paths) = files.get_mut(&filename) {
                    paths.push(f);
                } else {
                    files.insert(filename, vec![f]);
                }
            }
        }
    }

    if files.is_empty() {
        None
    } else {
        Some(files)
    }
}
